from browser import document, alert, html

#
# Base units converter (c) Greenfrogs
#

def _remove_duplicates(items):
    new_items = []
    for item in items:
        if item not in new_items:
            new_items.append(item)
    return new_items

def _sort_quick(units):
    order = ["C", "kg", "J", "Pa", "V", "W", "N", "F", "Ω", "S", "Wb", "T", "A", "m", "mol", "Hz", "K", "s", "cd", "lm", "lx", "sr", "kat"]
    units.sort(key=lambda i: order.index(i.partition("^")[0]))
    return units

def _sort_final(units):
    order = ["C", "kg", "J", "Pa", "V", "W", "N", "F", "Ω", "S", "Wb", "T", "A", "m", "mol", "Hz", "K", "s", "cd", "lm", "lx", "sr", "kat"]
    unitsA = []
    unitsB = []
    for unit in units:
        if "^-" in unit:
            unitsB.append(unit)
        else:
            unitsA.append(unit)
    unitsA.sort(key=lambda i: order.index(i.partition("^")[0]))
    unitsB.sort(key=lambda i: order.index(i.partition("^")[0]))
    unitsA.extend(unitsB)
    return unitsA

def print_beautify(units):
    output = ""
    for unit in units:
        if "^" in unit and document["latex"].checked:
            output = output + unit.split("^")[0] + "^{" + unit.split("^")[1] + "}⋅"
        else:
            output = output + unit + "⋅"
    return output.strip("⋅")

def _score_quick(units): #units consisting of an array e.g. ["s^-1", "s^-1"] = 2.1
    # unit^1 = 1, unit^-1 = 1.05
    score = 0.0
    for unit in units:
        if "^-" in unit:
            score += 1.05
        else:
            score += 1.0
    return score


def _score(units): #units consisting of an array e.g. ["s^-1", "s^-1"] = 2.1
    # unit^1 = 1, unit^-1 = 1.05
    score = 0.0
    for unit in units:
        if "^-" in unit:
            score += 1.05
        else:
            score += 1.0
        for unit2 in units:
            if unit2 == unit:
                score -= 0.0001
    return score # TODO: might want to make this more efficeint


def _contains_derived(units, base):
    for unit in units:
        if "^-1" in unit:
            if unit[:-3] not in base:
                return True
        else:
            if unit not in base:
                return True
    return False


def _convert_base(units, derived, base):
    keys = derived.keys()
    while _contains_derived(units, base):
        newUnits = []
        for unit in units:
            if unit in keys:
                newUnits.extend(derived[unit])
            else:
                newUnits.append(unit)
        units = newUnits
    return units


def _special_remove(units): # Removing sr as it is dimensionless, this is not going to cause any bugs later
    temp = []
    for unit in units:
        if "sr" not in unit:
            temp.append(unit)
    return temp

def _compress_units(units):
    compressed = False
    while not compressed:
        compressed = True
        for unit in units:
            if unit + "^-1" in units:
                units.remove(unit)
                units.remove(unit + "^-1")
                compressed = False
                break
    return units


def _split_powers(units):
    temp=[]
    for unit in units:
        if "^" in unit:
            if "-" in unit:
                for i in range(0, int(unit.split("^")[1][1:])):
                    temp.append(unit.split("^")[0] + "^-1")
            else:
                for i in range(0, int(unit.split("^")[1])):
                    temp.append(unit.split("^")[0])
        else:
            temp.append(unit)
    return temp


def _combine_powers(units):
    temp = []
    for unit in units:
        if "-" in unit:
            temp.append(str(unit.partition("^-")[0]) + "^-" + str(units.count(unit)))
        else:
            if units.count(unit) > 1:
                temp.append(str(unit) + "^" + str(units.count(unit)))
            else:
                temp.append(unit)
    return _remove_duplicates(temp)


def _combine(parent, child, derived, base): #This does not convert to base units first and just looks at what it has been given
    parent.append(child)
    child = _convert_base([child], derived, base)
    for unit in child:
        if unit in parent:
            parent.remove(unit)
        else: #Oppposite time, if s^-1 then s if s then s^-1
            if "^-1" in unit:
                parent.append(unit[:-3])
            else:
                parent.append(unit + "^-1")
    return parent


def _convert_input(user, derived, base): #Input can be in any form e.g. kg^2 S^2 or kg^1T^2
    keys = base.copy()
    keys.extend(derived.keys())
    keys.remove("Wb")
    keys.remove("mol")
    keys.append("Q")
    keys.append("Z")
    user = user.replace("O", "Ω") #There is an exception if Wb exists then W is found
    user = user.replace(" ", "")
    user = user.replace("Wb", "Q")
    user = user.replace("mol", "Z")
    lastpos = pos = 0
    out = []
    while pos < len(user):
        if user[lastpos:pos].split("^")[0] in keys:
            if user[pos:pos+1] == "^" or user[pos:pos+1] == "-" or user[pos:pos+1].isdigit():
                pos += 1
            else:
                out.append(user[lastpos:pos].replace("Q", "Wb").replace("Z", "mol"))
                lastpos = pos
        else:
            pos += 1
    out.append(user[lastpos:pos].replace("Q", "Wb").replace("Z", "mol"))
    return out


def _verify_input(user, derived, base):
    keys = base.copy()
    keys.extend(derived.keys())
    for unit in user:
        if unit.split("^")[0] not in keys:
            return False
    return True


def _find(units, base, derived, depth, discard, intel): #discard = if score greater than discard multiplied by base then give up with that root
    units = _compress_units(_convert_base(units, derived, base))
    directory = [units.copy()]
    currentDirectory = directory.copy()
    baseScore = _score(units)
    bestScore = baseScore * discard #If intel switched on then discard should be greater than 1
    keys = derived.keys()
    for i in range(0, depth):
        addDirectory = []
        for der in keys:
            for dir in currentDirectory:
                newUnits = _combine(dir.copy(), der, derived, base)
                score = _score_quick(newUnits)
                if score < bestScore:
                    addDirectory.append(_sort_quick(newUnits.copy()))
                    if intel and (bestScore > score * discard):
                        bestScore = score * discard
        newDirectory = []
        for i in addDirectory:
            if i not in currentDirectory and i not in directory:
                newDirectory.append(i)
                directory.append(i)
            elif i not in directory:
                directory.append(i)
        currentDirectory = newDirectory.copy()
    return directory



def unit_converter(ev):
    user = document["units"].value
    document["result"].html = '<div style="display: flex; justify-content: center;"><img src="images/loading.gif"></div>'
    base = ["kg", "m", "s", "A", "K", "mol", "cd", "sr"]
    derived = {
        "N": ["kg", "m", "s^-1", "s^-1"],
        "Pa":["kg", "m^-1", "s^-1", "s^-1"],
        "J": ["kg", "m", "m", "s^-1", "s^-1"],
        "W": ["kg", "m", "m", "s^-1", "s^-1", "s^-1"],
        "C": ["s", "A"],
        "V": ["kg", "m", "m", "s^-1", "s^-1", "s^-1", "A^-1"],
        "F": ["kg^-1", "m^-1", "m^-1", "s", "s", "s", "s", "A", "A"],
        "Ω": ["kg", "m", "m", "s^-1", "s^-1", "s^-1", "A^-1", "A^-1"],
        "S": ["kg^-1", "m^-1", "m^-1", "s", "s", "s", "A", "A"],
        "Wb": ["kg", "m", "m", "s^-1", "s^-1", "A^-1"],
        "T": ["kg", "s^-1", "s^-1", "A^-1"],
        "lm" : ["cd", "sr"],
        "lx": ["m^-1", "m^-1", "cd", "sr"],
        "kat": ["mol", "s^-1"]
    }
    user = _convert_input(user, derived, base)
    if _verify_input(user, derived, base):
        # Add derived^-1 into the derived list
        for key in derived.copy():
            inv = []
            for element in derived[key]:
                if "^-1" in element:
                    inv.append(element[:-3])
                else:
                    inv.append(element + "^-1")
            derived[key + "^-1"] = inv


        user = _split_powers(user)
        fail = False

        try:
            complexity = int(document["complexity"].value)
        except:
            complexity = 8
            document["result"].html = "Incorrect Complexity Value, Resetting"
            document["complexity"].value = 8
            fail = True

        try:
            discard = float(document["discard"].value)
        except:
            discard = 1.3
            document["result"].html = "Incorrect Discard Threshold, Resetting"
            document["discard"].value = 1.3
            fail = True
        if not fail:
            output = _find(user, base, derived, complexity, discard, True)
            output.sort(key=_score)

            out = '<table style="width:100%; text-align: center"><tr>'
            pos = 1
            for i in output:
                if document["latex"].checked:
                    out += '<td width="20%">\(' + print_beautify(_combine_powers(_special_remove(i))) + '\)</td>'
                else:
                    out += '<td width="20%">' + print_beautify(_combine_powers(_special_remove(i))) + '</td>'
                if pos > 3:
                    break
                pos += 1
            out = out + '</tr></table>'
            if document["latex"].checked:
                out = out + '<br>Base Units: \(' + print_beautify(_combine_powers(_compress_units(_special_remove(_convert_base(user, derived, base))))) + '\)'
            else:
                out = out + '<br>Base Units: ' + print_beautify(_combine_powers(_compress_units(_special_remove(_convert_base(user, derived, base))))) + ''
            document["result"].html = out
    else:
        document["result"].html = "Incorrect Format or Units"

document["button_unit_convert"].bind("click", unit_converter)